--
-- "Follow Me" - AutoFollow a Breadcrumb-trail
--
-- @team    Freelance Modding Crew (FMC)
-- @author  Decker_MMIV - fs-uk.com, forum.farming-simulator.com, modhoster.com
-- @date    2011-02-01
--          2012-08-29 (resumed)
--
-- Modifikationen erst nach Rcksprache
-- Do not edit without my permission
--
-- @history
--      v0.5(beta)  - Beta-release for internal FMC testing/criticism/suggestions
--      v0.51(beta) - Managed to get followers to use speed-throttle (well, not good enough), so they do
--                     not always drive as fast as possible.
--                  - Fixed problem of "circling a breadcrumb that vehicle can never reach", by calculating if
--                     vehicle is "in front of" the breadcrumb, which in that case it will go to the next one instead.
--      v0.52(beta) - Changed so the functions are appended to Steerable specialization, instead of Hirable, as there
--                     may be many "Truck"-mods that do not have Hirable.
--                  - Resets breadcrumb trail at every start. Previous version kept the trail "alive".
--      v0.53(beta) - TextBold=false, TextColor=white.
--  2012-November
--      v0.54       - Lots of changes. Attempt at FS2011 & FS2013 code.
--      v0.55       - Preliminary multiplayer support added.
--  2012-December
--      v0.56       - Misc. cleanup.
--                  - Added offset-left/right ability for follower (default keys: RIGHT CTRL + A / D)
--                  - Added keep back distance ability for follower (default keys: RIGHT CTRL + W / S)
--      v0.57       - Ability for the leading vehicle to "control" offset and keep-back on the follower,
--                    and only the _one_ following, _not_ the entire "chain".
--                      (default keys: RIGHT SHIFT + W / S and RIGHT SHIFT + A / D)
--      v0.58       - Read/write network stream functions changed.
--                  - Checks for existance of beaconlight, before attempting to turn on/off.
--      v0.59       - The 'steeringAngleLimit' for AIVehicleUtil.driveInDirection() should NOT be larger.
--      v0.60       - Misc. tweaks.
--                  - Attempt at translations to German, Polish & French.
--      v0.61       - If motor is not started, do not allow auto-follow. (Bug discovered by DrUptown)
--                  - Added ability to stop follower (default keys: RIGHT SHIFT + F)
--      v0.62       - Renamed input-binding to 'AutoFollowMeToggle', from 'AutoFollowDrive'.
--                  - Found that ESLimiter causes problems, due to its usage of self.isHired and setting speed-level.
--      v0.63       - Put all 'Follow Me' variables into a self.modFM, to reduce potential clashes with other mods/specializations.
--                  - Renamed/prefixed everything to 'FollowMe', instead of 'AutoFollow'.
--      v0.64       - More tweaking of the follower's speed.
--  2012-Xmas
--      v0.65       - Followers now duplicate the leader's dropped "crumbs", and will not make their own.
--                    This is to fix an "increasing amplitude" problem, when having a "convoy of vehicles".
--                    However, this also causes the "annoyance", that from a fresh loaded savegame each vehicle in
--                    a convoy must be driven a little distance first _before_ activating the FollowMe, so the
--                    vehicle behind it can "discover/locate" the trail.
--                  - For clients, added setStateFollowStalker() which hopefully will make the wheels turn etc. 
--                    when vehicle is in follow-mode (i.e. hired).
--                  - More tweaking of the follower's speed (this never ends).
--  2013-January
--      v0.66       - Changed some functionality with regards to network-messages between server and clients.
--      v0.67       - Changed minimum-distance-between-drops to 5 (from 10), and max-entries to 100 (from 50).
--  2013-February
--      v0.68       - Misc. cleanup.
--                  - Made some preliminary documentation for the mod.

--[[
Bugs/Suggestions from v0.52:

oliver45
    what do you do if you deleted the leader, because when i try making a new leader its says "there is already a leader" ?
    http://www.ls-uk.info/forum/index.php?topic=118980.msg767135#msg767135

MacPolska
    would it be possible to have some control over the distance and/or speed of the follower from the leader?
    http://www.ls-uk.info/forum/index.php?topic=118980.msg767298#msg767298

Feterlj
    it would be nice to have an offset function so that if the leader is using an implement that is centered with the tractor, but the follower is using an implement that is not centered with the tractor, you could offset him to do the job on the correct path.
    http://www.ls-uk.info/forum/index.php?topic=118980.msg767394#msg767394

Rsmasterrs
    put on beacons, they should stay on when you tell the vechicle to follow
    http://www.ls-uk.info/forum/index.php?topic=118980.msg767403#msg767403

rh
    - As someone mentioned above, an offset function
    - Beacons/lights for followers? Not sure if this is possible, but would be nice if say the beacons came on when the leader's beacons were on - you get the idea
    - Adjustable following distance?
    http://www.ls-uk.info/forum/index.php?topic=118980.msg767732#msg767732

seederman
    I have found when I tried a lead and 2 followers the 3rd one tends to get real close to the 2nd actually gets dirty and mounts 2nd lol, is there a way to spread them out so they don't get too close
    http://www.ls-uk.info/forum/index.php?topic=118980.msg768374#msg768374

Rondo
    is there in some way a possibility to alter the distance between the leader and the follower?
    http://www.ls-uk.info/forum/index.php?topic=118980.msg768652#msg768652
]]

--

FollowMe = {};

FollowMe.logLevel = 0; -- 1=Seldom occuring logs, 2=Event occuring logs, 3=Often occuring logs, 4=VeryOften occuring logs
FollowMe.debugDraw = {}

FollowMe.cMinDistanceBetweenDrops        =   5;   -- TODO, make configurable
FollowMe.cBreadcrumbsMaxEntries          = 100;   -- TODO, make configurable
FollowMe.cMstimeBetweenDrops             =  40;   -- TODO, make configurable

FollowMe.STATE_NONE      = 0;
FollowMe.STATE_TOGGLE    = 1;
FollowMe.STATE_START     = 2;
FollowMe.STATE_STOP      = 3;


function FollowMe:log(logLevel, txt)
  if (FollowMe.logLevel >= logLevel) then
    print(string.format("%7ums FollowMe(%s) ", g_currentMission.time, tostring(networkGetObjectId(self))) .. txt);
  end;
end;


-- FS2013
-- Support-function, that I would like to see be added to InputBinding class.
-- Maybe it is, I just do not know what its called.
function getKeyIdOfModifier(binding)
    if InputBinding.actions[binding] == nil then
        return nil;  -- Unknown input-binding.
    end;
    if table.getn(InputBinding.actions[binding].keys1) <= 1 then
        return nil; -- Input-binding has only one or zero keys. (Well, in the keys1 - I'm not checking keys2)
    end;
    -- Check if first key in key-sequence is a modifier key (LSHIFT/RSHIFT/LCTRL/RCTRL/LALT/RALT)
    if Input.keyIdIsModifier[ InputBinding.actions[binding].keys1[1] ] then
        return InputBinding.actions[binding].keys1[1]; -- Return the keyId of the modifier key
    end;
    return nil;
end
--]]

--[[ FS2011
--http://stackoverflow.com/questions/656199/search-for-an-item-in-a-lua-list
function Set(list)
    local set = {};
    for _,l in ipairs(list) do
        set[l]=true;
    end;
    return set;
end;

-- Support-function, that I would like to see be added to InputBinding class.
-- Maybe it is, I just do not know what its called.
function getKeyIdOfModifier(binding)
    local allowedModifiers = Set({
        Input.KEY_lshift,
        Input.KEY_rshift,
        Input.KEY_shift,
        Input.KEY_lctrl,
        Input.KEY_rctrl,
        Input.KEY_lalt,
        Input.KEY_ralt
    });
    for _,k in pairs(InputBinding.digitalActions[binding].key1Modifiers) do
        if allowedModifiers[k] then
            return k;
        end;
    end;
    return nil;
end;
--]]

function removeFromString(src, toRemove)
    local srcArr = Utils.splitString(" ", src);
    local remArr = Utils.splitString(" ", toRemove);
    local result = "";
    for i,p in ipairs(srcArr) do
        if i>1 then
            local found=false;
            for _,r in pairs(remArr) do
                if p == r then
                    found=true
                    break;
                end
            end
            if not found then
                result = result .. (result~="" and " " or "") .. p;
            end;
        end;
    end;
    return result;
end;

function FollowMe.initialize()
    if FollowMe.isInitialized then
        return;
    end;
    FollowMe.isInitialized = true;

    -- Get the modifier-key (if any) from input-binding
    FollowMe.keyModifier_FollowMeMyToggle = getKeyIdOfModifier(InputBinding.FollowMeMyToggle);

    -- Test that these four use the same modifier-key
        FollowMe.keyModifier_FollowMeMy  = getKeyIdOfModifier(InputBinding.FollowMeMyDistDec);
    if (FollowMe.keyModifier_FollowMeMy ~= getKeyIdOfModifier(InputBinding.FollowMeMyDistInc)
    or  FollowMe.keyModifier_FollowMeMy ~= getKeyIdOfModifier(InputBinding.FollowMeMyOffsDec)
    or  FollowMe.keyModifier_FollowMeMy ~= getKeyIdOfModifier(InputBinding.FollowMeMyOffsInc)
    ) then
        -- warning!
        FollowMe:log(0, "WARNING: Not all action-keys(1) use the same modifier-key!");
    end;

    -- Build a string, that is much shorter than what InputBinding.getKeyNamesOfDigitalAction() returns
    FollowMe.keys_FollowMeMy = FollowMe.keyModifier_FollowMeMy ~= nil and getKeyName(FollowMe.keyModifier_FollowMeMy) or "";
    FollowMe.keys_FollowMeMy = FollowMe.keys_FollowMeMy:upper();

    local shortKeys = removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeMyDistDec), FollowMe.keys_FollowMeMy)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeMyDistInc), FollowMe.keys_FollowMeMy)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeMyOffsDec), FollowMe.keys_FollowMeMy)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeMyOffsInc), FollowMe.keys_FollowMeMy);

    FollowMe.keys_FollowMeMy = FollowMe.keys_FollowMeMy .. " " .. shortKeys;

--print(FollowMe.keys_FollowMeMy);

    -- Test that these four use the same modifier-key
        FollowMe.keyModifier_FollowMeFl  = getKeyIdOfModifier(InputBinding.FollowMeFlDistDec);
    if (FollowMe.keyModifier_FollowMeFl ~= getKeyIdOfModifier(InputBinding.FollowMeFlDistInc)
    or  FollowMe.keyModifier_FollowMeFl ~= getKeyIdOfModifier(InputBinding.FollowMeFlOffsDec)
    or  FollowMe.keyModifier_FollowMeFl ~= getKeyIdOfModifier(InputBinding.FollowMeFlOffsInc)
    ) then
        -- warning!
        FollowMe:log(0, "WARNING: Not all action-keys(2) use the same modifier-key!");
    end;

    -- Build a string, that is much shorter than what InputBinding.getKeyNamesOfDigitalAction() returns
    FollowMe.keys_FollowMeFl = FollowMe.keyModifier_FollowMeFl ~= nil and getKeyName(FollowMe.keyModifier_FollowMeFl) or "";
    FollowMe.keys_FollowMeFl = FollowMe.keys_FollowMeFl:upper();

    local shortKeys = removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeFlDistDec), FollowMe.keys_FollowMeFl)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeFlDistInc), FollowMe.keys_FollowMeFl)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeFlOffsDec), FollowMe.keys_FollowMeFl)
            .. "," .. removeFromString(InputBinding.getKeyNamesOfDigitalAction(InputBinding.FollowMeFlOffsInc), FollowMe.keys_FollowMeFl);

    FollowMe.keys_FollowMeFl = FollowMe.keys_FollowMeFl .. " " .. shortKeys;

--print(FollowMe.keys_FollowMeFl);

end;

--
--
--

function FollowMe.load(self, xmlFile)
    --
    FollowMe.initialize();

    -- A simple attempt at making a "namespace" for 'Follow Me' variables.
    self.modFM = {};
    --
    self.modFM.IsInstalled = true;  -- TODO. Make 'FollowMe' a buyable add-on! This is expensive equipment ;-)
    --
    self.modFM.sumSpeed = 0;
    self.modFM.sumCount = 0;
    self.modFM.DropperCircularArray = {};
    self.modFM.DropperCurrentIndex = 0;
    self.modFM.StalkerVehicleObj = nil;  -- Needed in case self is being deleted.
    --
    self.modFM.FollowVehicleObj = nil;  -- What vehicle is this one following (if any)
    self.modFM.FollowCurrentIndex = 0;
    --self.modFM.lastZeroSpeedIndex = self.modFM.FollowCurrentIndex;
    self.modFM.FollowKeepBack = 0;
    self.modFM.FollowXOffset = 0;
    --
    self.modFM.ShowWarningText = nil;
    self.modFM.ShowWarningTime = 0;
    --
    self.modFM.isDirty = false;
    --
    if self.isServer then
        -- Copied from FS2011-Hirable, for the mods that do not include that specialization in their vehicle-type.
        self.modFM.PricePerMS = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.pricePerHour"), 2000)/60/60/1000;
        --
-- FS2013
        -- TODO: Await FS2013-scripts, so proper traffic-collision can be put in.
        --if self.addCollisionTrigger == nil then
        --    self.modFM.FollowKeepBack = 2;
        --    FollowMe:log(1, "WARNING: Seems there are no vehicle-collision-trigger(s) for this vehicle-type: ".. tostring(self.typeName));
        --end;
        if not (SpecializationUtil.hasSpecialization(AITractor, self.specializations)
             or SpecializationUtil.hasSpecialization(AICombine, self.specializations)) then
            self.modFM.FollowKeepBack = 2;
            FollowMe:log(1, "WARNING: Seems there are no vehicle-collision-trigger(s) for this vehicle-type: ".. tostring(self.typeName));
        end;
--]]
--[[ FS2011
        -- Copied from AITractor/AICombine, for mods that do not include that specialization in their vehicle-type.
        self.numCollidingVehicles = 0;
        if (self.aiTrafficCollisionTrigger == nil) then
            self.aiTrafficCollisionTrigger = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.aiTrafficCollisionTrigger#index"));
        end;
        if (self.aiTrafficCollisionTrigger == nil) then
            self.modFM.FollowKeepBack = 2;
            FollowMe:log(1, "WARNING: 'aiTrafficCollisionTrigger' is missing for this vehicle-type: ".. tostring(self.typeName));
        end;
--]]

        -- Drop one "crumb", to get it started...
        local wx,wy,wz = getWorldTranslation(self.components[1].node);
        FollowMe.addDrop(self, wx,wy,wz, 30/3600);
    end;
end;

function FollowMe.delete(self)
    if self.isServer then
        if self.modFM.StalkerVehicleObj ~= nil then
            -- Stop the stalker-vehicle
            FollowMe.stopFollowMe(self.modFM.StalkerVehicleObj, true);
        end;
        -- Stop ourself
        FollowMe.stopFollowMe(self, true);
    end;
    --
    --FollowMe.isInitialized = false;
end;

function FollowMe.sharedWriteStream(serverToClients, streamId, vehObj, followsObj, stalkedByObj, state, keepBackDistance, xOffset, warnTxt)
    local vehId = (networkGetObjectId(vehObj) or 0);

    streamWriteInt32(streamId, vehId);
    streamWriteUInt8(streamId, keepBackDistance);
    streamWriteInt8( streamId, xOffset);
    
    if (serverToClients) then
        -- server to clients
        local followsId     = (networkGetObjectId(followsObj)    or 0);
        local stalkedById   = (networkGetObjectId(stalkedByObj)  or 0);
        warnTxt = (warnTxt ~= nil and warnTxt or "");
        
        streamWriteInt32( streamId, followsId);
        streamWriteInt32( streamId, stalkedById);
        streamWriteString(streamId, warnTxt);
    else
        -- client to server
        streamWriteUInt8(streamId, state);
    end;
end;

function FollowMe.sharedReadStream(serverToClients, streamId)
    local state;
    local warnTxt;     
    local followsObj;  
    local stalkedByObj;
    
    local vehId            = streamReadInt32(streamId);
    local keepBackDistance = streamReadUInt8(streamId);
    local xOffset          = streamReadInt8( streamId);
    
    if (serverToClients) then
        -- server to clients
        local followsId;   
        local stalkedById; 
        
        followsId    = streamReadInt32( streamId);
        stalkedById  = streamReadInt32( streamId);
        warnTxt      = streamReadString(streamId);

        followsObj   = (followsId   ~= 0 and networkGetObject(followsId)   or nil);
        stalkedByObj = (stalkedById ~= 0 and networkGetObject(stalkedById) or nil);
    else
        -- client to server
        state        = streamReadUInt8(streamId);
    end;
    
    local vehObj = (vehId ~= 0 and networkGetObject(vehId) or nil);

    return
        vehObj,
        followsObj,
        stalkedByObj,
        state,
        keepBackDistance,
        xOffset,
        warnTxt;
end;

function FollowMe.writeStream(self, streamId, connection)
    FollowMe.sharedWriteStream(
        true,
        streamId,
        self,
        self.modFM.FollowVehicleObj,
        self.modFM.StalkerVehicleObj,
        FollowMe.STATE_NONE,
        self.modFM.FollowKeepBack,
        self.modFM.FollowXOffset,
        self.modFM.ShowWarningText
    );
end;

function FollowMe.readStream(self, streamId, connection)
    local dummySelf, dummyState, dummyWarnTxt;
    --
    dummySelf, 
    self.modFM.FollowVehicleObj, 
    self.modFM.StalkerVehicleObj, 
    dummyState, 
    self.modFM.FollowKeepBack, 
    self.modFM.FollowXOffset, 
    dummyWarnTxt                = FollowMe.sharedReadStream(true, streamId);
end;

function FollowMe.mouseEvent(self, posX, posY, isDown, isUp, button)
end;

function FollowMe.keyEvent(self, unicode, sym, modifier, isDown)
end;

function FollowMe.setWarning(self, txt, noSendEvent)
    self.modFM.ShowWarningText = txt;   -- must be a string that can be given to g_i18n:getText()
    self.modFM.ShowWarningTime = g_currentMission.time + 2500;
    --
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_NONE, txt, noSendEvent);
    if self.isServer then
        self.modFM.isDirty = true;
    end;
end;

function FollowMe.copyDrop(self, crumb)
    assert(g_server ~= nil);

    self.modFM.DropperCurrentIndex = self.modFM.DropperCurrentIndex + 1; -- Keep incrementing index, so followers will be able to detect if they get too far behind of the circular-array.
    local dropIndex = 1+((self.modFM.DropperCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries);
    
    self.modFM.DropperCircularArray[dropIndex] = crumb;
end;

function FollowMe.addDrop(self, wx,wy,wz, avgSpeed)
    assert(g_server ~= nil);

    self.modFM.DropperCurrentIndex = self.modFM.DropperCurrentIndex + 1; -- Keep incrementing index, so followers will be able to detect if they get too far behind of the circular-array.
    local dropIndex = 1+((self.modFM.DropperCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries);

    local rx,ry,rz  = localDirectionToWorld(self.components[1].node, 0,0,1);
    --local avgSpeed = math.max((self.modFM.sumSpeed / (self.modFM.sumCount>0 and self.modFM.sumCount or 1)), (5/3600));
    self.modFM.DropperCircularArray[dropIndex] = {
        trans = {wx,wy,wz},
        rot = {rx,ry,rz},
        avgSpeed = avgSpeed
        --acc = (acceleration ~= nil and acceleration or 1.0),
        --dropTime = g_currentMission.time
    };

    --FollowMe:log(3, string.format("Crumb #%d(%d): trans=%f/%f/%f, rot=%f/%f/%f, avgSpeed=%f, movTime=%f", FollowMe.gBreadcrumbsCurrentDropIndex,dropIndex, wx,wy,wz, rx,ry,rz, avgSpeed, self.modFM.movingTime));
end;

function FollowMe.changeDistance(self, newKeepBack, noSendEvent)
    local minKeepBack = 0;
-- FS2013    
    if not (SpecializationUtil.hasSpecialization(AITractor, self.specializations)
         or SpecializationUtil.hasSpecialization(AICombine, self.specializations)) then
        minKeepBack = 2; -- If no traffic-collision-trigger, keep back minimum 2 "crumbs".
    end;
--]]    
    local prevValue = self.modFM.FollowKeepBack;
    self.modFM.FollowKeepBack = Utils.clamp(newKeepBack, minKeepBack, math.floor(FollowMe.cBreadcrumbsMaxEntries / 2));
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_NONE, nil, noSendEvent);
    if prevValue ~= self.modFM.FollowKeepBack then
        self.modFM.isDirty = true;
    end;
end;

function FollowMe.changeXOffset(self, newXOffset, noSendEvent)
    local prevValue = self.modFM.FollowXOffset;
    self.modFM.FollowXOffset = Utils.clamp(newXOffset, -20.0, 20.0);
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_NONE, nil, noSendEvent);
    if prevValue ~= self.modFM.FollowXOffset then
        self.modFM.isDirty = true;
    end;
end;

function FollowMe.update(self, dt)
    if self:getIsActiveForInput(
-- FS2013
        false
--]]
       )
    then
        if InputBinding.hasEvent(InputBinding.FollowMeMyToggle) then
            FollowMe.toggleFollowMe(self);
        end;
        if self.modFM.FollowVehicleObj ~= nil then
            if     InputBinding.hasEvent(InputBinding.FollowMeMyDistDec) then  FollowMe.changeDistance(self, self.modFM.FollowKeepBack - 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeMyDistInc) then  FollowMe.changeDistance(self, self.modFM.FollowKeepBack + 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeMyOffsDec) then  FollowMe.changeXOffset(self, self.modFM.FollowXOffset - 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeMyOffsInc) then  FollowMe.changeXOffset(self, self.modFM.FollowXOffset + 1);
            end;
        end;
        if self.modFM.StalkerVehicleObj ~= nil then
            local stalker = self.modFM.StalkerVehicleObj;
            if     InputBinding.hasEvent(InputBinding.FollowMeFlStop)    then  FollowMe.toggleFollowMe(stalker);
            elseif InputBinding.hasEvent(InputBinding.FollowMeFlDistDec) then  FollowMe.changeDistance(stalker, stalker.modFM.FollowKeepBack - 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeFlDistInc) then  FollowMe.changeDistance(stalker, stalker.modFM.FollowKeepBack + 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeFlOffsDec) then  FollowMe.changeXOffset(stalker, stalker.modFM.FollowXOffset - 1);
            elseif InputBinding.hasEvent(InputBinding.FollowMeFlOffsInc) then  FollowMe.changeXOffset(stalker, stalker.modFM.FollowXOffset + 1);
            end;
        end;
    end;
    --
    --if (self.isServer and self.modFM.FollowVehicleObj ~= nil) then
    --    -- Copied from FS2011-Hirable
    --    local difficultyMultiplier = Utils.lerp(0.6, 1, (g_currentMission.missionStats.difficulty-1)/2) -- range from 0.6 (easy)  to  1 (hard)
    --    g_currentMission:addSharedMoney(-dt * difficultyMultiplier * self.modFM.PricePerMS);
    --end;
end;

function FollowMe.updateTick(self, dt)
  if self.isServer then
    --
    if self.modFM.IsInstalled then
      if self.modFM.FollowVehicleObj ~= nil then -- Have leading vehicle to follow.
        -- Simon Says: Beacon lights
        if self.beaconLightsActive ~= nil and self.modFM.FollowVehicleObj.beaconLightsActive ~= nil and self.modFM.FollowVehicleObj.beaconLightsActive ~= self.beaconLightsActive then
            self:setBeaconLightsVisibility(self.modFM.FollowVehicleObj.beaconLightsActive);
        end;
        --
        FollowMe.updateFollowMovement(self, dt);
        --
        -- Copied from FS2011-Hirable
        local difficultyMultiplier = Utils.lerp(0.6, 1, (g_currentMission.missionStats.difficulty-1)/2) -- range from 0.6 (easy)  to  1 (hard)
        g_currentMission:addSharedMoney(-dt * difficultyMultiplier * self.modFM.PricePerMS);
      elseif (self.movingDirection > 0) then  -- Must drive forward to drop crumbs
        self.modFM.sumSpeed = self.modFM.sumSpeed + self.lastSpeed;
        self.modFM.sumCount = self.modFM.sumCount + 1;
        --
        local wx,wy,wz = unpack(self.lastPosition); --getWorldTranslation(self.components[1].node); -- current position
        local pwx,pwy,pwz = unpack(self.modFM.DropperCircularArray[1+((self.modFM.DropperCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries)].trans); -- previous position
        local distancePrevDrop = Utils.vector2Length(pwx-wx, pwz-wz);
        if distancePrevDrop >= FollowMe.cMinDistanceBetweenDrops then
            local avgSpeed = math.max((self.modFM.sumSpeed / (self.modFM.sumCount>0 and self.modFM.sumCount or 1)), (5/3600));
            FollowMe.addDrop(self, wx,wy,wz, avgSpeed);
            --
            self.modFM.sumSpeed = 0;
            self.modFM.sumCount = 0;
        end;
      end;
    end;
  end;
  --
  FollowMe.sendUpdate(self);
end;

function FollowMe:sendUpdate(stateId)
  if self.modFM.isDirty or stateId ~= nil then
    self.modFM.isDirty = false;
    --
    if self.isServer then
      -- Remove warning-text if not needed anymore
      if self.modFM.ShowWarningTime < g_currentMission.time then
        self.modFM.ShowWarningText = nil;
      end;
      -- Broadcast current state to all clients.
      FollowMeEvent.sendEvent(self, FollowMe.STATE_NONE);
    else
      -- Only send the client's action-commands to server.
      FollowMeEvent.sendEvent(self, stateId~=nil and stateId or FollowMe.STATE_NONE);
    end;
  end;
end;

function FollowMe:recvUpdate(stateId, keepBackDist, xOffset, followsObj, stalkedByObj, warnTxt)
  if self.isServer then
    -- Received a client's action-commands.
    FollowMe.changeDistance(self, keepBackDist);
    FollowMe.changeXOffset(self, xOffset);
    if stateId ~= FollowMe.STATE_NONE then
      FollowMe.toggleFollowMe(self, stateId);
    end;
    -- If anything was changed (made dirty), the next updateTick() will broadcast to all clients
  else
    -- Received the server's state.
    local prevDirty = self.modFM.isDirty;
    
    FollowMe.changeDistance(self, keepBackDist);
    FollowMe.changeXOffset(self, xOffset);
    if warnTxt ~= nil and warnTxt ~= "" then
      FollowMe.setWarning(self, warnTxt)
    end;
    FollowMe.setStateFollowStalker(self, followsObj, stalkedByObj);

    self.modFM.isDirty = prevDirty;
  end;
end;

function FollowMe.updateFollowMovement(self, dt)
    assert(self.modFM.FollowVehicleObj ~= nil);

    -- Bug-fix: If motor is turned off, stop following.
    if not self.isMotorStarted then
        FollowMe.stopFollowMe(self);
        return;
    end;
    
    local allowedToDrive = true;
    local moveForwards = true;
    --
-- FS2013
    if self.numCollidingVehicles ~= nil then
        for _,numCollisions in pairs(self.numCollidingVehicles) do
            if numCollisions > 0 then
                allowedToDrive = false;
                break;
            end;
        end;
    end
--]]
--[[ FS2011
    if self.numCollidingVehicles ~= nil and self.numCollidingVehicles > 0 then
        allowedToDrive = false;
    end;
--]]
    --
    local acceleration = self.lastRealAcceleration;
    local crumb;
    local tx,ty,tz;
    if allowedToDrive then
        local leader = self.modFM.FollowVehicleObj;
        if self.modFM.FollowCurrentIndex == leader.modFM.DropperCurrentIndex then
            -- do nothing, waiting for another "crumb"
            --FollowMe:log(2, "Waiting for next drop-index.");
            allowedToDrive = false;
        elseif self.modFM.FollowCurrentIndex <= (leader.modFM.DropperCurrentIndex - FollowMe.cBreadcrumbsMaxEntries) then
            -- circular-array have "circled" once, and this follower did not move fast enough.
            --FollowMe:log(2, "Much too far behind. Stopping auto-follow.");
            FollowMe.setWarning(self, "FollowMeTooFarBehind");
            FollowMe.stopFollowMe(self);
            return;
        else
            if (leader.modFM.DropperCurrentIndex - self.modFM.FollowCurrentIndex < self.modFM.FollowKeepBack) then
                allowedToDrive = false;
            elseif (leader.modFM.DropperCurrentIndex - self.modFM.FollowCurrentIndex < self.modFM.FollowKeepBack+2) then
                -- Approaching leading vehicle, so slow down
                --acceleration = 0.8;
            end;
            --
            crumb = leader.modFM.DropperCircularArray[1+((self.modFM.FollowCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries)];
            if crumb ~= nil then
                local cx,cy,cz = getWorldTranslation(self.components[1].node);  -- current location
                local crx,cry,crz  = localDirectionToWorld(self.components[1].node, 0,0,1); -- current rotation
                --
                tx,ty,tz = crumb.trans[1],crumb.trans[2],crumb.trans[3]; -- target location
                local trx,try,trz = unpack(crumb.rot); -- target rotation
--
                -- Apply offset to target-point
                tx = tx - trz * self.modFM.FollowXOffset;
                tz = tz + trx * self.modFM.FollowXOffset;
                --
                local dx,dz = tx - cx, tz - cz;
                local dist = Utils.vector2Length(dx,dz);
                -- Rotate to see if the point is still "in front of us"
                local rAngle = math.atan2(trx,trz);
              --local nx = dx * math.cos(rAngle) - dz * math.sin(rAngle);
                local nz = dx * math.sin(rAngle) + dz * math.cos(rAngle);
                --FollowMe:log(4, string.format("Follow: dist=%f, dxdz=%f/%f, rot=%f/%f/%f, angle=%f, nxnz=%f/%f", dist, dx,dz, rx,ry,rz, rAngle, nx,nz));
                if (dist < 4)   -- closer than 4 units from crumb?
                or (nz < 0)     -- in front of crumb?
                then
                    FollowMe.copyDrop(self, crumb);
                    -- Go to next crumb
                    self.modFM.FollowCurrentIndex = self.modFM.FollowCurrentIndex + 1;
                    crumb = nil; -- Wait to next updateTick()
                    --crumb = leader.modFM.DropperCircularArray[1+((self.modFM.FollowCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries)];
                    --FollowMe:log(3, string.format("Next crumb:%d(%d)", self.modFM.FollowCurrentIndex, 1+((self.modFM.FollowCurrentIndex-1) % FollowMe.cBreadcrumbsMaxEntries)));
                elseif dist > 50 then
                    --FollowMe:log(2, "Too far away from next waypoint. Stopping auto-follow.");
                    FollowMe.setWarning(self, "FollowMeTooFarBehind");
                    FollowMe.stopFollowMe(self);
                    return;
                end;
            else
                -- Hmm? Not supposed to happen.
                FollowMe.setWarning(self, "FollowMeTooFarBehind");
                FollowMe.stopFollowMe(self);
                return;
            end;
        end;
    end;
    --
    if not allowedToDrive then
        -- TODO - Try to make a not-so-strong emergency brake.
        AIVehicleUtil.driveInDirection(self, dt, 30, 0.0, 0.0, 28, allowedToDrive, moveForwards, 0.0,1.0, nil, 1);
        ---- If basically stopped...
        --if self.lastSpeed < 0.0002 then
        --    self.modFM.lastZeroSpeedIndex = self.modFM.FollowCurrentIndex + 1;
        --end;
    elseif crumb ~= nil then
        local lx,lz = AIVehicleUtil.getDriveDirection(self.components[1].node, tx,ty,tz);

        -- TODO - How to determine what speed the vehicle should have? It is actually not that easy apparently;
        --        1) If leader drives slowly or halts, the follower should not slow down nor halt when it reaches the same "crumb".
        --        2) The follower should not drive faster than the leader... but what is the max/average speed of the leader anyway?

        ----if (leader.modFM.DropperCurrentIndex - self.modFM.FollowCurrentIndex < self.modFM.FollowKeepBack+4) then
        --if (g_currentMission.time - crumb.dropTime < 5000) then
        --    -- We're less than 5 seconds behind, so try to adjust the speed
        --    local speedDiff = math.max((crumb.avgSpeed - (self.lastSpeed * 0.9)), 0);
        --    acceleration = math.min(speedDiff / (crumb.avgSpeed * 0.20), acceleration);
        --else
        --    -- We're 5 seconds or more behind, step on the pedal!
        --end;
        ----acceleration = crumb.acc;
        
        
        local speedDiffPct = (self.lastSpeed - crumb.avgSpeed) --/ crumb.avgSpeed;
        speedDiffPct = (speedDiffPct*3600) / (crumb.avgSpeed*3600);
        --if speedDiffPct > 0.2 then
        --    acceleration = 1.0 -- Step on it!
        --else
        --    acceleration = Utils.clamp(acceleration + (acceleration * speedDiffPct), 0.0, 1.0);
        --end;
        if speedDiffPct < -0.20 -- below required averaged speed
        --or (self.modFM.FollowCurrentIndex <= self.modFM.lastZeroSpeedIndex) -- if previously have been totally stopped.
        then
            acceleration = 1.0;
        elseif speedDiffPct > 0.20 then -- above required average speed
            acceleration = 0.5; --  A bit too fast for some vehicles, and a bit too slow for others.
        else
            acceleration = math.min(acceleration + (acceleration * (-speedDiffPct / 2)), 1.0);
        end;

--[[DEBUG
local dbgId = tostring(networkGetObjectId(self));
FollowMe.debugDraw[dbgId.."a1"] = {"last/avg",     string.format("%f - %f  (%.1f - %.1f)", self.lastSpeed, crumb.avgSpeed, self.lastSpeed*3600, crumb.avgSpeed*3600)}
FollowMe.debugDraw[dbgId.."a3"] = {"diffSpeedPct", string.format("%.2f", speedDiffPct)}
FollowMe.debugDraw[dbgId.."a4"] = {"acceleration", string.format("%.1f", acceleration)}

--FollowMe.debugDraw[dbgId.."b1"] = {"drive-vector", string.format("%f,%f", lx,lz)}

FollowMe.debugDraw[dbgId.."z9"] = {"", ""}
--DEBUG ]]
        --
        self.motor.maxRpmOverride = nil; -- 
        AIVehicleUtil.driveInDirection(self, dt, 30, acceleration, (acceleration * 0.8), 30, allowedToDrive, moveForwards, lx,lz, nil, 1);
    end;
end;

function FollowMe.toggleFollowMe(self, stateId, noSendEvent)
    if not self.modFM.IsInstalled then
        FollowMe.setWarning(self, "FollowMeNotAvailable");
    elseif self.isServer then
--[[DEBUG
FollowMe.debugDraw = {};
--DEBUG ]]
        if self.modFM.FollowVehicleObj == nil and (stateId == nil or stateId ~= FollowMe.STATE_STOP) then
            FollowMe.startFollowMe(self, noSendEvent);
        else
            FollowMe.stopFollowMe(self, noSendEvent);
        end;
    else
        --FollowMeEvent.sendEvent(self, FollowMe.STATE_TOGGLE, nil, noSendEvent);
        FollowMe.sendUpdate(self, FollowMe.STATE_TOGGLE);
    end;
end;

function FollowMe.setStalker(self, stalkerVeh, noSendEvent)
    self.modFM.StalkerVehicleObj = stalkerVeh;
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_NONE, nil, noSendEvent);
    self.modFM.isDirty = self.isServer and true or self.modFM.isDirty;
end;

function FollowMe.setStateFollowStalker(self, followsObj, stalkedByObj)
    assert(g_server == nil);    -- Only for clients
    --
    self.modFM.FollowVehicleObj  = followsObj;
    self.modFM.StalkerVehicleObj = stalkedByObj;
    --
    -- Try to fix the problem of clients not seeing wheel-rotation and the "farmer-in-the-seat".
    if (self.modFM.FollowVehicleObj ~= nil) then
        self.forceIsActive = true;
        self.stopMotorOnLeave = false;
        self.steeringEnabled = false;
        self.deactivateOnLeave = false;
        self.disableCharacterOnLeave = false;
    else
        self.forceIsActive = false;
        self.stopMotorOnLeave = true;
        self.steeringEnabled = true;
        self.deactivateOnLeave = true;
        self.disableCharacterOnLeave = true;
        if not self.isEntered and not self.isControlled then
            if self.characterNode ~= nil then
                setVisibility(self.characterNode, false);
            end;
        end;
    end;
end;

function FollowMe.startFollowMe(self, noEventSend)
    assert(g_server ~= nil);

--[[ FS2011
    if self.aiTrafficCollisionTrigger == nil then
        FollowMe.setWarning(self, "FollowMeMissingTrafficCollisionTrigger");
        return;
    end;
--]]

    if self.modFM.FollowVehicleObj ~= nil then
        --FollowMe.setWarning(self, "FollowMeActive");
        return;
    end;
    -- Make sure the motor is turned on, before allowing to auto-follow
    if not self.isMotorStarted then
        FollowMe.setWarning(self, "FollowMeStartEngine");
        return;
    end;

    --
    local wx,wy,wz = getWorldTranslation(self.components[1].node);
    local rx,ry,rz = localDirectionToWorld(self.components[1].node, 0,0,1);
    local rlength = Utils.vector2Length(rx,rz);
    local rotDeg = math.deg(math.atan2(rx/rlength,rz/rlength));
    local rotRad = Utils.degToRad(rotDeg-45.0);
    local rotRad = Utils.degToRad(rotDeg-45.0);
    --FollowMe:log(2, string.format("getWorldTranslation:%f/%f/%f - localDirectionToWorld:%f/%f/%f - rDeg:%f - rRad:%f", wx,wy,wz, rx,ry,rz, rotDeg, rotRad));

    -- Find closest vehicle, that is in front of self.
    local closestDistance = 50;
    local closestVehicle = nil;
    for _,vehicleObj in pairs(g_currentMission.steerables) do
        if vehicleObj.modFM.DropperCircularArray ~= nil -- Make sure other vehicle has circular array
        and vehicleObj.modFM.StalkerVehicleObj == nil then -- and is not already stalked by something.
            local vx,vy,vz = getWorldTranslation(vehicleObj.components[1].node);
            local dx,dz = vx-wx, vz-wz;
            local dist = Utils.vector2Length(dx,dz);
            if (dist < closestDistance) then
                -- Rotate to see if vehicleObj is "in front of us"
                local nx = dx * math.cos(rotRad) - dz * math.sin(rotRad);
                local nz = dx * math.sin(rotRad) + dz * math.cos(rotRad);
                if (nx > 0) and (nz > 0) then
                    closestDistance = dist;
                    closestVehicle = vehicleObj;
                end;
            end;
        end;
    end;

    if closestVehicle == nil then
        FollowMe.setWarning(self, "FollowMeDropperNotFound");
        return;
    end;

    -- Find closest "breadcrumb"
    self.modFM.FollowCurrentIndex = 0;
    local closestDistance = 50;
    for i=closestVehicle.modFM.DropperCurrentIndex, math.max(closestVehicle.modFM.DropperCurrentIndex - FollowMe.cBreadcrumbsMaxEntries,1), -1 do
        local crumb = closestVehicle.modFM.DropperCircularArray[1+((i-1) % FollowMe.cBreadcrumbsMaxEntries)];
        if crumb ~= nil then
            local x,y,z = unpack(crumb.trans);
            -- Translate
            local dx,dz = x-wx, z-wz;
            local dist = Utils.vector2Length(dx,dz);
            --local r = Utils.getYRotationFromDirection(dx,dz);
            --FollowMe:log(3, string.format("#%d - xz:%f/%f - dxdz:%f/%f - r:%f - dist:%f", i, x,z, dx,dz, r, dist));
            if (dist > 2) and (dist < closestDistance) then
                -- Rotate to see if the point is "in front of us"
                local nx = dx * math.cos(rotRad) - dz * math.sin(rotRad);
                local nz = dx * math.sin(rotRad) + dz * math.cos(rotRad);
                if (nx > 0) and (nz > 0) then
                    --FollowMe:log(3, string.format("#%d - xz:%f/%f - dxdz:%f/%f - dist:%f - nxnz:%f/%f", i, x,z, dx,dz, dist, nx,nz));
                    closestDistance = dist;
                    self.modFM.FollowCurrentIndex = i;
                end;
            end;
        end;
    end;
    --FollowMe:log(2, string.format("ClosestDist:%f, index:%d", closestDistance, self.modFM.FollowCurrentIndex));
    --
    if self.modFM.FollowCurrentIndex == 0 then
        self.modFM.FollowVehicleObj = nil;
        FollowMe.setWarning(self, "FollowMeDropperNotFound");
        return;
    end;

    --
    --self.modFM.lastZeroSpeedIndex = self.modFM.FollowCurrentIndex;
    
    -- Chain with leading vehicle.
    self.modFM.FollowVehicleObj = closestVehicle;
    FollowMe.setStalker(self.modFM.FollowVehicleObj, self);

    --
-- FS2013
    if SpecializationUtil.hasSpecialization(AITractor, self.specializations) then
        AITractor.addCollisionTrigger(self, self);
    elseif SpecializationUtil.hasSpecialization(AICombine, self.specializations) then
        AICombine.addCollisionTrigger(self, self);
    else
        -- TODO - Display warning!
    end;
--]]
--[[ FS2011
    self.numCollidingVehicles = 0;
    if (self.onTrafficCollisionTrigger ~= nil) then
        addTrigger(self.aiTrafficCollisionTrigger, "onTrafficCollisionTrigger", self);
    else
        addTrigger(self.aiTrafficCollisionTrigger, "FollowMeOnTrafficCollisionTrigger", self);
    end;
--]]
    --
--[[ ESLimiter causes problem, when using Hirable.
    if (self.hire ~= nil) then
        -- Use the FS2011-Hirable specialization's functionality
        self:hire();
    else
--]]    
        -- Copied from FS2011-Hirable, for the mods that do not include that specialization in their vehicle-type.
        self.forceIsActive = true;
        self.stopMotorOnLeave = false;
        self.steeringEnabled = false;
        self.deactivateOnLeave = false;
        self.disableCharacterOnLeave = false;
--  end;
    --
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_START, noEventSend);
    self.modFM.isDirty = true;
end;

function FollowMe.stopFollowMe(self, noSendEvent)
    assert(g_server ~= nil);

    if self.modFM.FollowVehicleObj == nil then
        return;
    end;

    -- Unchain with leading vehicle.
    assert(self.modFM.FollowVehicleObj.modFM.StalkerVehicleObj == self);
    FollowMe.setStalker(self.modFM.FollowVehicleObj, nil);
    --
    self.modFM.FollowVehicleObj = nil;
    self.modFM.FollowCurrentIndex = 0;

    --
--[[ ESLimiter causes problem, when using Hirable.
    if (self.dismiss ~= nil) then
        -- Use the FS2011-Hirable specialization's functionality
        self:dismiss();
    else
--]]    
      -- Copied from FS2011-Hirable, for the mods that do not include that specialization in their vehicle-type.
      self.forceIsActive = false;
      self.stopMotorOnLeave = true;
      self.steeringEnabled = true;
      self.deactivateOnLeave = true;
      self.disableCharacterOnLeave = true;
      if not self.isEntered and not self.isControlled then
          if self.characterNode ~= nil then
              setVisibility(self.characterNode, false);
          end;
      end;
--  end;

    self.motor:setSpeedLevel(0, false);
    self.motor.maxRpmOverride = nil;

    WheelsUtil.updateWheelsPhysics(self, 0, self.lastSpeed, 0, false, self.requiredDriveMode);

-- FS2013
    if SpecializationUtil.hasSpecialization(AITractor, self.specializations) then
        AITractor.removeCollisionTrigger(self, self);
    elseif SpecializationUtil.hasSpecialization(AICombine, self.specializations) then
        AICombine.removeCollisionTrigger(self, self);
    end;
--]]
--[[ FS2011
    if self.aiTrafficCollisionTrigger ~= nil then
        removeTrigger(self.aiTrafficCollisionTrigger);
    end;
--]]
    --
    --FollowMeEvent.sendEvent(self, FollowMe.STATE_STOP, noSendEvent);
    self.modFM.isDirty = true;
end;

--[[ FS2011
-- Copied from AITractor. Used in case the vehicle-type does not have AITractor or AICombine.
function FollowMe.FollowMeOnTrafficCollisionTrigger(self, triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if otherId == g_currentMission.player.rootNode then
            if onEnter then
                self.numCollidingVehicles = self.numCollidingVehicles+1;
            elseif onLeave then
                self.numCollidingVehicles = math.max(self.numCollidingVehicles-1, 0);
            end;
        else
            local vehicle = g_currentMission.nodeToVehicle[otherId];
            if vehicle ~= nil
            --and self.trafficCollisionIgnoreList[otherId] == nil
            then
                if onEnter then
                    self.numCollidingVehicles = self.numCollidingVehicles+1;
                elseif onLeave then
                    self.numCollidingVehicles = math.max(self.numCollidingVehicles-1, 0);
                end;
            end;
        end;
    end;
end;
--]]

function FollowMe.draw(self)
    if self.modFM.ShowWarningTime > g_currentMission.time then
        g_currentMission:addWarning(g_i18n:getText(self.modFM.ShowWarningText), 0.07+0.022, 0.019+0.029);
    end;
    --
    if g_currentMission.showHelpText then
        if self.modFM.FollowVehicleObj ~= nil 
        or FollowMe.keyModifier_FollowMeMyToggle == nil 
        or (FollowMe.keyModifier_FollowMeMyToggle ~= nil and Input.isKeyPressed(FollowMe.keyModifier_FollowMeMyToggle)) then
            g_currentMission:addHelpButtonText(g_i18n:getText("FollowMeMyToggle"), InputBinding.FollowMeMyToggle);
        end;
        --
        if self.modFM.FollowVehicleObj ~= nil then
            g_currentMission:addExtraPrintText(
                string.format(
                    g_i18n:getText("HelpKeysMyself")
                    ,FollowMe.keys_FollowMeMy
                    ,self.modFM.FollowKeepBack
                    ,self.modFM.FollowXOffset
                )
            );
        end;
        --
        if self.modFM.StalkerVehicleObj ~= nil then
            g_currentMission:addHelpButtonText(g_i18n:getText("FollowMeFlStop"), InputBinding.FollowMeFlStop);
            g_currentMission:addExtraPrintText(
                string.format(
                    g_i18n:getText("HelpKeysBehind")
                    ,FollowMe.keys_FollowMeFl
                    ,self.modFM.StalkerVehicleObj.modFM.FollowKeepBack
                    ,self.modFM.StalkerVehicleObj.modFM.FollowXOffset
                )
            );
        end;
--[[ DEBUG        
    else
        --if self.modFM.FollowVehicleObj ~= nil then
            local yPos = 0.9;
            setTextColor(1,1,1,1);
            setTextBold(true);
            local keys = {}
            for k,_ in pairs(FollowMe.debugDraw) do
                table.insert(keys,k);
            end;
            table.sort(keys);
            for _,k in pairs(keys) do
                local v = FollowMe.debugDraw[k];
                yPos = yPos - 0.02;
                renderText(0.01, yPos, 0.02, v[1]);
                renderText(0.11, yPos, 0.02, v[2]);
            end;
            setTextBold(false);
        --end;
-- DEBUG]]
    end;
end;

--
Steerable.load        = Utils.appendedFunction( Steerable.load,          FollowMe.load);
Steerable.delete      = Utils.appendedFunction( Steerable.delete,        FollowMe.delete);
Steerable.writeStream = Utils.appendedFunction( Steerable.writeStream,   FollowMe.writeStream);
Steerable.readStream  = Utils.appendedFunction( Steerable.readStream,    FollowMe.readStream);
Steerable.update      = Utils.prependedFunction(Steerable.update,        FollowMe.update);
Steerable.updateTick  = Utils.appendedFunction( Steerable.updateTick,    FollowMe.updateTick);
Steerable.draw        = Utils.appendedFunction( Steerable.draw,          FollowMe.draw);

---
---
---

FollowMeEvent = {};
FollowMeEvent_mt = Class(FollowMeEvent, Event);

InitEventClass(FollowMeEvent, "FollowMeEvent");

function FollowMeEvent:emptyNew()
    local self = Event:new(FollowMeEvent_mt);
    self.className = "FollowMeEvent";
    return self;
end;

function FollowMeEvent:new(vehicle, stateId)
    local self = FollowMeEvent:emptyNew()
    self.vehicle      = vehicle;
    self.stateId      = stateId;
    return self;
end;

function FollowMeEvent:writeStream(streamId, connection)
    FollowMe.sharedWriteStream(
        g_server ~= nil,
        streamId, 
        self.vehicle, 
        self.vehicle.modFM.FollowVehicleObj, 
        self.vehicle.modFM.StalkerVehicleObj, 
        self.stateId, 
        self.vehicle.modFM.FollowKeepBack, 
        self.vehicle.modFM.FollowXOffset, 
        self.vehicle.modFM.ShowWarningText
    );
end;

function FollowMeEvent:readStream(streamId, connection)
    local vehObj, followsObj, stalkedByObj, stateId, keepBackDist, xOffset, warnTxt;
    vehObj, 
    followsObj, 
    stalkedByObj, 
    stateId, 
    keepBackDist, 
    xOffset, 
    warnTxt         = FollowMe.sharedReadStream(g_server == nil, streamId);
    --
    if vehObj ~= nil then
        FollowMe.recvUpdate(vehObj, stateId, keepBackDist, xOffset, followsObj, stalkedByObj, warnTxt);
    end;
end;

function FollowMeEvent.sendEvent(vehicle, stateId, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(FollowMeEvent:new(vehicle, stateId), nil, nil, vehicle);
        else
            g_client:getServerConnection():sendEvent(FollowMeEvent:new(vehicle, stateId));
        end;
    end;
end;


--
--function FollowMe.getIsHired(self, superFunc)
--    if self.modFM ~= nil and self.modFM.FollowVehicleObj ~= nil then
--        return true;
--    end;
--    return superFunc(self);
--end;
--
--Utils.overwrittenFunction(Vehicle.getIsHired, FollowMe.getIsHired)

--
print("Script loaded: FollowMe.lua (v0.68)");
